@@ -3,7 +3,6 @@ and defines various simple operations on it. *)
 open Pdfutil
 open Pdfio
 
-
 (* Predicate on characters delimiting entities. *)
 let is_delimiter = function
   | '(' | ')' | '<' | '>' | '[' | ']' | '{' | '}' | '%' | '/' -> true
@@ -146,7 +145,6 @@ type pdfobjects =
    mutable pdfobjects : pdfobjmap;
    mutable object_stream_ids : (int, int) Hashtbl.t}
 
-
 (* PDF Document. The major and minor version numbers, the root object number,
 the list of objects and the trailer dictionary.
 
@@ -163,8 +161,8 @@ type t =
 
 (* The null PDF document. *)
 let empty () =
-  {major = 1;
-   minor = 1;
+  {major = 2;
+   minor = 0;
    root = 0;
    objects =
      {maxobjnum = 0;
@@ -445,7 +443,7 @@ let lookup_exception (exp : exn) pdf key dict =
   let dict' =
     match direct pdf dict with
     | Dictionary d | Stream {contents = Dictionary d, _} -> d
-    | _ -> raise (PDFError ("lookup_exception: not a dictionary, key = " ^ key))
+    | _ -> raise exp
   in
     match lookup key dict' with
     | None | Some Null -> raise exp
@@ -931,7 +929,9 @@ let rec contents_of_nametree pdf tree =
   | Some (Array names) ->
       let rec pairs_of_list prev = function
       | [] -> rev prev
-      | [_] -> raise (PDFError "contents_of_nametree: bad /Names")
+      | [_] ->
+          Pdfe.log "warning: contents_of_nametree: odd number of /Names\n";
+          rev prev
       | k::v::r -> pairs_of_list ((k, v)::prev) r
       in
         pairs_of_list [] names
@@ -995,3 +995,39 @@ let change_id pdf f =
       pdf.trailerdict <-
         Dictionary (add "/ID" (generate_id pdf f (fun () -> Random.float 1.)) d)
   | _ -> raise (PDFError "Bad trailer dictionary")
+
+let transform_rect pdf transform rect =
+  let minx, miny, maxx, maxy = parse_rectangle pdf rect in
+    let (x0, y0) = Pdftransform.transform_matrix transform (minx, miny) in
+    let (x1, y1) = Pdftransform.transform_matrix transform (maxx, maxy) in
+    let (x2, y2) = Pdftransform.transform_matrix transform (minx, maxy) in
+    let (x3, y3) = Pdftransform.transform_matrix transform (maxx, miny) in
+      let minx = fmin (fmin x0 x1) (fmin x2 x3) in
+      let miny = fmin (fmin y0 y1) (fmin y2 y3) in
+      let maxx = fmax (fmax x0 x1) (fmax x2 x3) in
+      let maxy = fmax (fmax y0 y1) (fmax y2 y3) in
+        Array [Real minx; Real miny; Real maxx; Real maxy]
+
+let transform_quadpoint_single pdf transform = function
+  | [x1; y1; x2; y2; x3; y3; x4; y4] ->
+      let x1, y1, x2, y2, x3, y3, x4, y4 =
+        getnum pdf x1, getnum pdf y1,
+        getnum pdf x2, getnum pdf y2,
+        getnum pdf x3, getnum pdf y3,
+        getnum pdf x4, getnum pdf y4
+      in
+        let (x1, y1) = Pdftransform.transform_matrix transform (x1, y1) in
+        let (x2, y2) = Pdftransform.transform_matrix transform (x2, y2) in
+        let (x3, y3) = Pdftransform.transform_matrix transform (x3, y3) in
+        let (x4, y4) = Pdftransform.transform_matrix transform (x4, y4) in
+          map (fun x -> Real x) [x1; y1; x2; y2; x3; y3; x4; y4]
+  | qp ->
+     Pdfe.log "Malformed /QuadPoints format: must be a multiple of 8 entries\n";
+     qp
+
+let transform_quadpoints pdf transform = function
+| Array qps ->
+    Array (flatten (map (transform_quadpoint_single pdf transform) (splitinto 8 qps)))
+| qp ->
+    Pdfe.log (Printf.sprintf "Unknown or malformed /QuadPoints format %s\n" (!string_of_pdf qp));
+    qp
