@@ -7,17 +7,13 @@ let debug = ref false
 let flate_level = ref 6
 
 (* Get the next non-whitespace character in a stream. *)
-let rec get_streamchar skipped s =
+let rec get_streamchar s =
   match s.input_byte () with
-  | x when x = Pdfio.no_more -> raise End_of_file
+  | x when x = Pdfio.no_more ->
+      raise End_of_file
   | x ->
-    let chr = Char.unsafe_chr x in
-      if Pdf.is_whitespace chr then
-        begin
-          incr skipped;
-          get_streamchar skipped s
-        end
-      else chr
+      let chr = Char.unsafe_chr x in
+        if Pdf.is_whitespace chr then get_streamchar s else chr
 
 (* Raised if there was bad data. *)
 exception Couldn'tDecodeStream of string
@@ -51,8 +47,8 @@ let decode_ASCIIHex i =
   in let enddata = ref false in
     try
       while not !enddata do
-        let b = get_streamchar (ref 0) i in
-          let b' = get_streamchar (ref 0) i in
+        let b = get_streamchar i in
+          let b' = get_streamchar i in
             match b, b' with
             | '>', _ ->
                 (* Rewind for inline images *)
@@ -87,25 +83,16 @@ let decode_ASCIIHex i =
           raise (Couldn'tDecodeStream "ASCIIHex")
 
 (* ASCII85 *)
-
-(* Decode five characters. *)
 let decode_5bytes c1 c2 c3 c4 c5 n o =
-  let d x p =
-    i32mul (i32ofi (int_of_char x - 33)) (i32ofi (pow p 85))
-  in
-    let total =
-      i32add
-        (i32add (d c1 4) (d c2 3))
-        (i32add (d c3 2) (i32add (d c4 1) (d c5 0)))
-    in
-      let extract t =
-        char_of_int (i32toi (lsr32 (lsl32 total (24 - t)) 24))
-      in
-        if n = 4 then extract 0::extract 8::extract 16::extract 24::o
-        else if n = 3 then extract 8::extract 16::extract 24::o
-        else if n = 2 then extract 16::extract 24::o
-        else if n = 1 then extract 24::o
-        else o
+  let d x p = i32mul (i32ofi (int_of_char x - 33)) (i32ofi (pow p 85)) in
+  let total = i32add (i32add (d c1 4) (d c2 3)) (i32add (d c3 2) (i32add (d c4 1) (d c5 0))) in
+  let extract t = char_of_int (i32toi (lsr32 (lsl32 total (24 - t)) 24)) in
+    match n with
+    | 4 -> extract 0::extract 8::extract 16::extract 24::o
+    | 3 -> extract 8::extract 16::extract 24::o
+    | 2 -> extract 16::extract 24::o
+    | 1 -> extract 24::o
+    | _ -> o
 
 let conso cs o =
   match cs with
@@ -116,33 +103,17 @@ let conso cs o =
   | _ -> o
 
 let rec decode_ASCII85_i i cs o =
-  match i.input_byte () with
-  | x when x = Pdfio.no_more -> raise End_of_file
-  | x ->
-     match char_of_int x with
-     | c when Pdf.is_whitespace c ->
-       if length cs = 5
-         then decode_ASCII85_i i [] (conso (rev cs) o)
-         else decode_ASCII85_i i cs o
-     | 'z' ->
-       decode_ASCII85_i i [] ('\000'::'\000'::'\000'::'\000'::conso (rev cs) o)
-     | '~' ->
-       begin match i.input_byte () with
-       | x when x = Pdfio.no_more -> raise End_of_file
-       | x ->
-          match char_of_int x with
-          | '>' ->
-             bytes_of_charlist (rev (conso (rev cs) o))
-          | c ->
-             if length cs = 5
-               then decode_ASCII85_i i [c] (conso (rev cs) o)
-               else decode_ASCII85_i i (c::cs) o
-       end
-     | c when c >= '!' && c <= 'u' ->
-       if length cs = 5
-         then (decode_ASCII85_i i [c] (conso (rev cs) o))
-         else (decode_ASCII85_i i (c::cs) o)
-     | _ -> raise (Failure "A")
+  match get_streamchar i with
+  | 'z' ->
+      decode_ASCII85_i i [] ('\000'::'\000'::'\000'::'\000'::conso (rev cs) o)
+  | '~' ->
+      bytes_of_charlist (rev (conso (rev cs) o))
+  | c when c >= '!' && c <= 'u' ->
+      if length cs = 5
+        then (decode_ASCII85_i i [c] (conso (rev cs) o))
+        else (decode_ASCII85_i i (c::cs) o)
+  | _ ->
+      raise (Pdf.PDFError "decode_ASCII85_i")
 
 let decode_ASCII85 i =
   try decode_ASCII85_i i [] [] with
@@ -151,29 +122,24 @@ let decode_ASCII85 i =
 (* Encode a single symbol set. *)
 let encode_4bytes = function
   | [b1; b2; b3; b4] ->
-      let ( * ) = Int64.mul
-      in let ( - ) = Int64.sub
-      in let ( / ) = Int64.div 
-      in let rem = Int64.rem in
+      let ( * ), ( - ), ( / ), rem = Int64.mul, Int64.sub, Int64.div, Int64.rem in
         let numbers =
           [i64ofi (int_of_char b1) * i64ofi (pow 3 256);
            i64ofi (int_of_char b2) * i64ofi (pow 2 256);
            i64ofi (int_of_char b3) * i64ofi (pow 1 256);
            i64ofi (int_of_char b4) * i64ofi (pow 0 256)]
         in
-          let t = fold_left Int64.add Int64.zero numbers
-          in let one85 = i64ofi (pow 1 85) in let two85 = i64ofi (pow 2 85)
-          in let three85 = i64ofi (pow 3 85) in let zero85 = i64ofi (pow 0 85)
-          in let four85 = i64ofi (pow 4 85) in
-            let t, c5 = t - rem t one85, rem t one85 / zero85 in
-              let t, c4 = t - rem t two85, rem t two85 / one85 in
-                let t, c3 = t - rem t three85, rem t three85 / two85 in
-                  let t, c2 = t - rem t four85, rem t four85 / three85 in
-                    let i1, i2, i3, i4, i5 =
-                      i64toi (t / four85),
-                      i64toi c2, i64toi c3, i64toi c4, i64toi c5
-                    in
-                      i1, i2, i3, i4, i5
+          let t = fold_left Int64.add Int64.zero numbers in
+          let one85 = i64ofi (pow 1 85) in
+          let two85 = i64ofi (pow 2 85) in
+          let three85 = i64ofi (pow 3 85) in
+          let zero85 = i64ofi (pow 0 85) in
+          let four85 = i64ofi (pow 4 85) in
+          let t, c5 = t - rem t one85, rem t one85 / zero85 in
+          let t, c4 = t - rem t two85, rem t two85 / one85 in
+          let t, c3 = t - rem t three85, rem t three85 / two85 in
+          let t, c2 = t - rem t four85, rem t four85 / three85 in
+            i64toi (t / four85), i64toi c2, i64toi c3, i64toi c4, i64toi c5
   | _ -> assert false
 
 (* Encode a stream. *)
@@ -297,19 +263,19 @@ let encode_flate stream =
 let debug_stream_serial = ref 0
 
 let debug_stream s =
-  (*Printf.eprintf "First 50 bytes%!\n";
+  Pdfe.log "First 50 bytes\n";
   for x = 0 to 50 do
-    Printf.eprintf "%C = %i\n%!" (char_of_int (bget s x)) (bget s x)
-  done*)
-  (* Write stream to current directory as <length>_<serial>.zlib *)
+    Pdfe.log (Printf.sprintf "%C = %i\n" (char_of_int (bget s x)) (bget s x))
+  done
+  (*(* Write stream to current directory as <length>_<serial>.zlib *)
   let name = string_of_int (bytes_size s) ^ "_" ^ string_of_int !debug_stream_serial ^ ".zlib" in
-  Printf.eprintf "Writing %s\n%!" name;
+    Pdfe.log (Printf.sprintf "Writing %s\n" name);
     debug_stream_serial += 1;
     let fh = open_out_bin name in
       for x = 0 to bytes_size s - 1 do
         output_byte fh (bget s x)
       done;
-      close_out fh
+      close_out fh*)
 
 let decode_flate stream =
   if bytes_size stream = 0 then mkbytes 0 else (* Accept the empty stream. *)
@@ -896,8 +862,7 @@ too low. Fails in the usual manner if the index is too high. *)
 let get0 a i =
   if i < 0 then 0 else a.(i)
 
-(* TIFF prediction. 8bpp only for now. We don't have example files for the
-others yet, so we'll have to wait to create our own examples. *)
+(* TIFF prediction. 8bpp only for now. *)
 let decode_tiff_predictor colors bpc columns stream =
   match bpc with
   | 1 -> raise (DecodeNotSupported "TIFF Predictor for 1bpc not supported")
@@ -1113,7 +1078,7 @@ let decode_runlength i =
         done
       with
         End_of_file ->
-          Printf.eprintf "Warning: Missing EOD marker in runlength decode...\n%!"
+          Pdfe.log "Warning: Missing EOD marker in runlength decode...\n"
       end;
       extract_bytes_from_input_output o data
 
@@ -1122,6 +1087,9 @@ type source =
   | StreamSource of bytes
   | InputSource of input
 
+let decode_identity i =
+  Pdfio.bytes_of_input i 0 i.in_channel_length
+
 let decoder pdf dict source name =
   let input_of_source = function
     | InputSource i -> i
@@ -1129,6 +1097,7 @@ let decoder pdf dict source name =
   in
     let i = input_of_source source in
       match name with
+      | "/Crypt" -> decode_identity i
       | "/ASCIIHexDecode" | "/AHx" -> decode_ASCIIHex i
       | "/ASCII85Decode" | "/A85" -> decode_ASCII85 i
       | "/FlateDecode" | "/Fl" ->
@@ -1288,8 +1257,7 @@ let decode_pdfstream_onestage pdf stream =
       begin match
         Pdf.direct pdf (Pdf.lookup_fail "no /Length" pdf "/Length" dict)
       with
-      | Pdf.Integer _ -> ()
-      (*i if l <> bytes_size s then raise (PDFError "Wrong /Length") i*)
+      | Pdf.Integer _ -> () (*i if l <> bytes_size s then raise (PDFError "Wrong /Length") i*)
       | _ -> raise (Pdf.PDFError "No /Length")
       end;
       let stream' = decode_one pdf dict (StreamSource s) in
@@ -1340,10 +1308,6 @@ let decode_from_input i dict =
         let rec decode_rest stream = function
           | [] -> stream
           | Pdf.Name _::more ->
-              (* Removing the filter just done, and its decodeparms. Note that
-              /F can denote a file specification, but this can never be in an
-              inline image (which is currrently the only use of
-              decode_from_input, so that's ok for now. *)
               let filters =
                 Pdf.lookup_direct_orelse (Pdf.empty ()) "/F" "/Filter" dict
               in let decodeparms =
@@ -1422,7 +1386,7 @@ let encoder_of_encoding = function
 let process_prediction_data predictor predictor_columns d =
   match predictor with
     PNGUp -> encode_predictor 12 1 8 predictor_columns d
-  | _ -> raise (Pdf.PDFError "process_predction_data")
+  | _ -> raise (Pdf.PDFError "process_prediction_data")
 
 let process_prediction predictor predictor_columns stream =
   match stream with
