@@ -126,7 +126,7 @@ let bytes_of_charlist cs =
         s
 
 let bytes_of_arraylist l =
-  let totalsize = fold_left ( + ) 0 (map Array.length l) in
+  let totalsize = sum (map Array.length l) in
     let output = mkbytes totalsize
     in let pos = ref 0 in
       iter
@@ -145,6 +145,7 @@ let string_of_bytes s =
       done;
       Buffer.contents buf
 
+
 let bytes_of_int_array a =
   let s = mkbytes (Array.length a) in
     for i = 0 to bytes_size s - 1 do
@@ -173,7 +174,7 @@ let int_array_of_string s =
   Array.init (String.length s) (fun i -> int_of_char (String.unsafe_get s i))
 
 let string_of_int_arrays arrays =
-  let len = fold_left ( + ) 0 (map Array.length arrays) in
+  let len = sum (map Array.length arrays) in
     let buf = Buffer.create len in
       iter (Array.iter (fun v -> Buffer.add_char buf (Char.unsafe_chr v))) arrays;
       Buffer.contents buf
@@ -418,9 +419,13 @@ let bytes_of_input_channel ch =
 
 (* Save a bytes to a channel. *)
 let bytes_to_output_channel ch data =
-  for x = 1 to bytes_size data do
-    output_byte ch (bget_unsafe data (x - 1))
-  done
+  match data with
+  | Short caml_bytes ->
+      output_string ch (Obj.magic caml_bytes : string)
+  | Long _ as data ->
+      for x = 1 to bytes_size data do
+        output_byte ch (bget_unsafe data (x - 1))
+      done
 
 (* Like Stdlib.read_line *) 
 let b = Buffer.create 256
@@ -488,6 +493,9 @@ let bytes_of_input i o l =
     else
       bytes_of_caml_bytes s
 
+let string_of_input i =
+  string_of_bytes (bytes_of_input i 0 i.in_channel_length)
+
 let getinit i s o l =
   if l = 0 then () else
   let max = bytes_size s - 1
@@ -674,11 +682,11 @@ let write_bitstream_append_aligned a b =
 let debug_next_char i =
   try
     let a = unopt (i.input_char ()) in
-      Printf.eprintf "%C = %i\n%!" a (int_of_char a)
+      Pdfe.log (Printf.sprintf "%C = %i\n" a (int_of_char a))
   with
     _ -> ()
 
 let debug_next_n_chars n i =
   for x = 1 to n do debug_next_char i done;
-  prerr_string "\n";
+  Pdfe.log "\n";
   for x = 1 to n do rewind i done
