@@ -236,19 +236,29 @@ let rec filterspecial = function
   | Pdf.Name ("/ASCIIHexDecode" | "/ASCII85Decode" | "/AHx" | "/A85")::_ -> true
   | _::t -> filterspecial t
 
+let b = Buffer.create 256
+
 let string_of_lexeme = function
   | LexComment -> ""
   | Obj o -> Pdfread.string_of_lexeme o 
   | Op op -> op
   | PdfObj obj -> Pdfwrite.string_of_pdf obj
   | LexInlineImage (dict, data) ->
+      (* Compress if no compression *)
+      let dict, data =
+        match Pdf.lookup_direct_orelse (Pdf.empty ()) "/F" "/Filter" dict with
+        | None | Some (Pdf.Array []) ->
+            Pdf.add_dict_entry dict "/F" (Pdf.Name "/Fl"),
+            Pdfcodec.encode_flate data
+        | _ -> dict, data
+      in
+      let dict = Pdf.add_dict_entry dict "/L" (Pdf.Integer (bytes_size data)) in
       let dict_string = Pdfwrite.string_of_pdf dict in
         let dict_string' =
           (* Remove the dictionary markers. *)
           implode (rev (drop' 2 (rev (drop' 2 (explode dict_string)))))
-        in let data_string =
-          string_of_bytes data
-        in let space =
+        in
+        let space =
           let filters =
             match
               Pdf.lookup_direct_orelse (Pdf.empty ()) "/F" "/Filter" dict
@@ -259,7 +269,9 @@ let string_of_lexeme = function
           in
             if filterspecial filters then "" else " "
         in
-          "BI\n" ^ dict_string' ^ " ID" ^ space ^ data_string ^ "\nEI\n"
+          Buffer.clear b;
+          iter (Buffer.add_string b) ["BI\n"; dict_string'; " ID"; space; string_of_bytes data; "\nEI\n"];
+          Buffer.contents b
 
 let b = Buffer.create 30
 
@@ -299,7 +311,7 @@ in the case of a malformed operator stream, not on any legitimate content. *)
 exception Couldn'tHandleContent
 
 let nocontent i =
-  Printf.eprintf "Failed to understand content on page\n%!";
+  Pdfe.log "Failed to understand content on page\n";
   if !Pdfread.read_debug then
     begin
       Pdfio.debug_next_n_chars 20 i;
@@ -354,12 +366,12 @@ let rec components pdf resources t =
   | Pdf.Array [Pdf.Name "/Pattern"; alternate] ->
       components pdf resources (Pdf.direct pdf alternate)
   | cs ->
-     Printf.eprintf "%s\n%!" (Pdfwrite.string_of_pdf cs);
+     Pdfe.log (Printf.sprintf "%s\n" (Pdfwrite.string_of_pdf cs));
      raise (Pdf.PDFError "Unknown colourspace")
 
 (* Lex an inline image. We read the dictionary, and then the stream. *)
 let lex_inline_image pdf resources i =
-  if !debug then Printf.eprintf "lex_inline_image at %i\n%!" (i.pos_in ());
+  if !debug then Pdfe.log (Printf.sprintf "lex_inline_image at %i\n" (i.pos_in ()));
   try
   let dict =
     let lexemes = Pdfread.lex_dictionary i in
@@ -367,7 +379,7 @@ let lex_inline_image pdf resources i =
         (Pdfread.parse
           ([Pdfgenlex.LexLeftDict] @ lexemes @ [Pdfgenlex.LexRightDict]))
   in
-    if !debug then Printf.eprintf "dict was %s\n%!" (Pdfwrite.string_of_pdf dict);
+    if !debug then Pdfe.log (Printf.sprintf "dict was %s\n" (Pdfwrite.string_of_pdf dict));
     (* Read ID token *)
     Pdfread.dropwhite i;
     let c = char_of_int (i.input_byte ()) in
@@ -385,7 +397,7 @@ let lex_inline_image pdf resources i =
           not (filterspecial filters)
       in
         if toskip then ignore (i.input_byte ());
-        if !debug then Printf.eprintf "**got ID header, skipped possble byte%!";
+        if !debug then Pdfe.log "**got ID header, skipped possble byte";
         let bytes =
           let bpc =
             match
@@ -393,7 +405,7 @@ let lex_inline_image pdf resources i =
             with
             | Some (Pdf.Integer bpc) -> bpc
             | _ ->
-                Printf.eprintf "no BPC\n%!";
+                Pdfe.log "no BPC\n";
                 nocontent i
           in
           let cspace =
@@ -404,17 +416,17 @@ let lex_inline_image pdf resources i =
             | Some (Pdf.Name ("/G" | "/RGB" | "/CMYK") as n) -> n
             | Some ((Pdf.Array _) as n) -> n
             | Some (Pdf.Name cspace) ->
-                if !debug then Printf.eprintf "resources is %s\n%!" (Pdfwrite.string_of_pdf resources);
+                if !debug then Pdfe.log (Printf.sprintf "resources is %s\n" (Pdfwrite.string_of_pdf resources));
                 begin match Pdf.lookup_direct pdf "/ColorSpace" resources with
                 | Some (Pdf.Dictionary _ as d) ->
                     begin match Pdf.lookup_direct pdf cspace d with
                     | Some c -> c
                     | _ ->
-                        Printf.eprintf "no colourspace A\n%!";
+                        Pdfe.log "no colourspace A\n";
                         nocontent i
                     end
                 | _ ->
-                    Printf.eprintf "no colourspace B\n%!";
+                    Pdfe.log "no colourspace B\n";
                     nocontent i
                 end
             | None ->
@@ -424,23 +436,23 @@ let lex_inline_image pdf resources i =
                 with
                 | Some (Pdf.Boolean true) -> Pdf.Name "/DeviceGray"
                 | _ ->
-                    Printf.eprintf "no colourspace C\n%!";
+                    Pdfe.log "no colourspace C\n";
                     nocontent i
                 end
             | _ ->
-                Printf.eprintf "no colourspace D\n%!";
+                Pdfe.log "no colourspace D\n";
                 nocontent i
           in let width =
             match Pdf.lookup_direct_orelse pdf "/W" "/Width" dict with
             | Some (Pdf.Integer w) -> w
             | _ ->
-                Printf.eprintf "no or malformed /W%!";
+                Pdfe.log "no or malformed /W";
                 nocontent i 
           in let height =
             match Pdf.lookup_direct_orelse pdf "/H" "/Height" dict with
             | Some (Pdf.Integer h) -> h
             | _ ->
-                Printf.eprintf "no or malformed /H%!";
+                Pdfe.log "no or malformed /H";
                 nocontent i
           in
             let bitwidth =
@@ -463,25 +475,25 @@ let lex_inline_image pdf resources i =
                     done;
                   data
                 with
-                | e -> Printf.eprintf "%s%!" (Printexc.to_string e); raise e
+                | e -> Pdfe.log (Printexc.to_string e); raise e
                 end
             | Some (Pdf.Name ("/DCT" | "/DCTDecode") | Pdf.Array [Pdf.Name ("/DCT" | "/DCTDecode")]) ->
                 (* FIXME. The case of DCT combined with another one is possible e.g ["/DCT"; "/A85"]. Need to re-work. But have not seen an example yet. *)
                 begin try Pdfjpeg.get_jpeg_data i with
                   e ->
-                    Printf.eprintf "Couldn't read inline image JPEG data %s\n%!" (Printexc.to_string e);
+                    Pdfe.log (Printf.sprintf "Couldn't read inline image JPEG data %s\n" (Printexc.to_string e));
                     raise e
                 end
             | Some _ ->
                 try
                   match Pdfcodec.decode_from_input i dict with
                   | None ->
-                      Printf.eprintf "decode_from_input failed\n%!";
+                      Pdfe.log "decode_from_input failed\n";
                       nocontent i
                   | Some data -> data 
                 with
                   | Pdfcodec.DecodeNotSupported d ->
-                      Printf.eprintf "Content DecodeNotSupported: %s\n%!" d;
+                      Pdfe.log (Printf.sprintf "Content DecodeNotSupported: %s\n" d);
                       nocontent i
                   | Pdfcodec.Couldn'tDecodeStream r ->
                       raise (Pdf.PDFError ("Inline image, bad data: " ^ r))
@@ -493,10 +505,15 @@ let lex_inline_image pdf resources i =
               let c' = char_of_int (i.input_byte ()) in
                 if c <> 'E' || c' <> 'I' then
                   begin
-                    Printf.eprintf "warning: bad end to inline image %C, %C\n%!" c c';
-                    (* We try to find "EI" anyway, in case there is just some junk.
-                     * To do this, we drop any whitespace and any E or I character. *)
-                    Pdfread.ignoreuntil true (fun x -> Pdf.is_not_whitespace x && x <> 'E' && x <> 'I') i;
+                    Pdfe.log (Printf.sprintf "warning: bad end to inline image %C, %C\n" c c');
+                    (* We try to find "EI" anyway, in case there is just some junk. This is
+                    basic - if the junk contains E....I somewhere in it, the procedure will fail. *)
+                    Pdfread.ignoreuntil true (( = ) 'E') i;
+                    Pdfread.ignoreuntil true (( = ) 'I') i;
+                    Pdfio.nudge i;
+                    Pdfread.dropwhite i;
+                    (* 29/03/2023: Previous, worse implementation here: *)
+                    (*Pdfread.ignoreuntil true (fun x -> Pdf.is_not_whitespace x && x <> 'E' && x <> 'I') i;*)
                   end;
                 (* Remove filter, predictor, if it wasn't JPEG. *)
                 let dict' =
@@ -514,11 +531,11 @@ let lex_inline_image pdf resources i =
                 in
                   dict', data
     | _ ->
-        Printf.eprintf "Did not recognise beginning of inline image ID\n%!";
+        Pdfe.log "Did not recognise beginning of inline image ID\n";
         nocontent i
   with
     e ->
-      Printf.eprintf "inline image reading failed: %s\n%!" (Printexc.to_string e);
+      Pdfe.log (Printf.sprintf "inline image reading failed: %s\n" (Printexc.to_string e));
       nocontent i
 
 (* Lex a keyword. *)
@@ -770,7 +787,7 @@ let parse_operator compatibility = function
                        | (Obj (Pdfgenlex.LexReal i)) -> Some (Pdf.Real i)
                        | (Obj (Pdfgenlex.LexString s)) -> Some (Pdf.String s)
                        | e ->
-                           Printf.eprintf "Warning: malformed TJ element; skipping\n%!";
+                           Pdfe.log "Warning: malformed TJ element; skipping\n";
                            None)
                       t
                   in
@@ -778,10 +795,10 @@ let parse_operator compatibility = function
               | _ -> raise (Pdf.PDFError "malformed TJ op")
               end
           | Op _::_ as l ->
-             Printf.eprintf "Empty or malformed graphics operation %s.\n%!" (string_of_lexemes (rev l));
+             Pdfe.log (Printf.sprintf "Empty or malformed graphics operation %s.\n" (string_of_lexemes (rev l)));
              Op_Unknown (string_of_lexemes (rev l))
           | l ->
-             Printf.eprintf "Empty or malformed graphics operation %s.\n%!" (string_of_lexemes (rev l));
+             Pdfe.log (Printf.sprintf "Empty or malformed graphics operation %s.\n" (string_of_lexemes (rev l)));
              Op_Unknown (string_of_lexemes (rev l))
       in
         r, more
@@ -798,7 +815,7 @@ split over several streams, which must be concatenated at the lexeme level. *)
 
 (* Concatenate bytess, padding with whitespace *)
 let concat_bytess ss =
-  let total_length = fold_left ( + ) 0 (map bytes_size ss) in
+  let total_length = sum (map bytes_size ss) in
     let s' = mkbytes (total_length + length ss) in
       let p = ref 0 in
         iter
@@ -812,12 +829,15 @@ let concat_bytess ss =
           ss;
         s'
 
-let parse_stream pdf resources streams =
-  let stream = match streams with [s] -> s | _ -> concat_bytess streams in
+let parse_single_stream pdf resources stream =
   let i = input_of_bytes stream in
     let lexemes = lex_stream pdf resources i in
       parse_lexemes (ref 0) lexemes []
 
+let parse_stream pdf resources streams =
+  let stream = match streams with [s] -> s | _ -> concat_bytess streams in
+    parse_single_stream pdf resources stream
+
 (* Parse the operators in a list of streams. *)
 let parse_operators pdf resources streams =
   let rawstreams =
