@@ -79,11 +79,19 @@ type t =
   | Op_EX
   | Op_Unknown of string
 
-(** Parse a single byte stream to an operator list given a document and resource
+(** Parse a single byte streams to an operator list given a document and resource
+dictionary. *)
+val parse_single_stream :
+  Pdf.t -> Pdf.pdfobject -> Pdfio.bytes -> t list
+
+(** Parse a list of byte streams to an operator list given a document and resource
 dictionary. *)
 val parse_stream :
   Pdf.t -> Pdf.pdfobject -> Pdfio.bytes list -> t list
 
+(** Concatenate a list bytes, padding with whitespace between. *)
+val concat_bytess : Pdfio.bytes list -> Pdfio.bytes
+
 (** Given a pdf document, resource dictionary and list of streams representing
 the graphics content (PDF allows a single page's graphics content to be split
 over several streams), return a list of operators. Raises PDFError on bad
