@@ -13,22 +13,18 @@ containing the graphical content stream (see the Pdfops module), mediabox
 the page size, resources the page's resource dictionary, rotate its rotation
 and rest any other entries to reside in the page dictionary. *)
 type t =
-  {content : Pdf.pdfobject list; (*FIXME Change this to int list, to ensure sharing? *)
+  {content : Pdf.pdfobject list;
    mediabox : Pdf.pdfobject;
    resources : Pdf.pdfobject;
    rotate : rotation;
-   rest : Pdf.pdfobject} (* A dictionary of the other records in the page. *)
+   rest : Pdf.pdfobject} (* A dictionary of the other entries in the page dictionary. *)
 
 (* Make a PDF rectangle from a Paper.papersize. *)
 let rectangle_of_paper paper =
-  let u = Pdfpaper.unit paper
-  in let w = Pdfpaper.width paper
-  in let h = Pdfpaper.height paper in
-    let w', h' =
-      let f = Pdfunits.convert 100. u Pdfunits.PdfPoint in
-        f w, f h
-    in
-      Pdf.Array [Pdf.Real 0.; Pdf.Real 0.; Pdf.Real w'; Pdf.Real h']
+  let u = Pdfpaper.unit paper in
+  let w = Pdfunits.points (Pdfpaper.width paper) u in
+  let h = Pdfunits.points (Pdfpaper.height paper) u in
+    Pdf.Array [Pdf.Real 0.; Pdf.Real 0.; Pdf.Real w; Pdf.Real h]
 
 (* Create a page with empty content, media box from the given paper size,
 empty resources, zero rotation and no extra dictionary entries. *)
@@ -79,6 +75,7 @@ let last_mediabox_seen =
   ref (Pdf.Array [Pdf.Integer 0; Pdf.Integer 0; Pdf.Integer 612; Pdf.Integer 792])
 
 let rec find_pages pages pdf resources mediabox rotate =
+  match pages with Pdf.Null -> [] | _ ->
   match Pdf.lookup_direct pdf "/Type" pages with
   | Some (Pdf.Name "/Pages") | None ->
       begin match
@@ -91,6 +88,7 @@ let rec find_pages pages pdf resources mediabox rotate =
                | Pdf.Indirect k ->
                    (try Pdf.lookup_obj pdf k with
                      Not_found -> raise (Pdf.PDFError "missing kid\n"))
+               | Pdf.Dictionary d -> Pdf.Dictionary d (* malformed but we allow *)
                | _ -> raise (Pdf.PDFError "malformed kid\n"))
               kids
           in
@@ -113,7 +111,7 @@ let rec find_pages pages pdf resources mediabox rotate =
                   kids)
       | _ -> raise (Pdf.PDFError "Malformed /Kids in page tree node")
       end
-  | Some (Pdf.Name "/Page") ->
+  | Some _ ->
       let resources =
         match Pdf.lookup_direct pdf "/Resources" pages with
         | Some x -> Some x
@@ -140,7 +138,7 @@ let rec find_pages pages pdf resources mediabox rotate =
                 (* An array of indirects. Just return it *)
                 Pdf.lookup_direct pdf "/Contents" pages
             end
-        | _ -> assert false
+        | _ -> raise (Pdf.PDFError "Pages not a dictionary")
         end
       in let rotate =
         match Pdf.lookup_direct pdf "/Rotate" pages with
@@ -174,7 +172,7 @@ let rec find_pages pages pdf resources mediabox rotate =
             (match mediabox with
             | Some m -> last_mediabox_seen := m; m
             | None ->
-                Printf.eprintf "Warning: missing mediabox. Using most recently seen.\n%!";
+                Pdfe.log "Warning: missing mediabox. Using most recently seen.\n";
                 !last_mediabox_seen);
           rotate = rotate;
           rest =
@@ -182,7 +180,6 @@ let rec find_pages pages pdf resources mediabox rotate =
             | Pdf.Dictionary d -> Pdf.Dictionary (remove_dict_entries [] d)
             | _ -> raise (Pdf.PDFError "Bad /Pages"))
         }]
-  | _ -> raise (Pdf.PDFError "find_pages: Not a page tree node or page object")
 
 (* Given a pdf, return a list of (resources, contents, mediabox) triples. *)
 let pages_of_pagetree pdf =
@@ -196,6 +193,7 @@ let pages_of_pagetree pdf =
       find_pages pages pdf None None Rotate0
 
 let rec find_pages_quick pages pdf =
+  match pages with Pdf.Null -> 0 | _ ->
   match Pdf.lookup_direct pdf "/Type" pages with
   | Some (Pdf.Name "/Pages") | None ->
       begin match
@@ -208,14 +206,14 @@ let rec find_pages_quick pages pdf =
                | Pdf.Indirect k ->
                    (try Pdf.lookup_obj pdf k with
                      Not_found -> raise (Pdf.PDFError "missing kid\n"))
-               | _ -> raise (Pdf.PDFError "malformed kid\n"))
+               | Pdf.Dictionary d -> Pdf.Dictionary d (* malformed, but we allow *)
+               | x -> raise (Pdf.PDFError "malformed kid\n"))
               kids
           in
-            fold_left ( + ) 0 (map (fun k -> find_pages_quick k pdf) kids)
+            sum (map (fun k -> find_pages_quick k pdf) kids)
       | _ -> raise (Pdf.PDFError "Malformed /Kids in page tree node")
       end
-  | Some (Pdf.Name "/Page") -> 1
-  | _ -> raise (Pdf.PDFError "find_pages: Not a page tree node or page object")
+  | Some _ -> 1
 
 let pages_of_pagetree_quick pdf =
   let document_catalog =
@@ -230,6 +228,9 @@ let pages_of_pagetree_quick pdf =
 let endpage pdf =
   pages_of_pagetree_quick pdf
 
+let _ =
+  Pdfread.endpage := endpage
+
 (* Make a collection of pages capable of being merged -- in other words rename
 their resources so as not to clash. *)
 let source k =
@@ -294,7 +295,7 @@ let change_operator pdf lookup lookup_option seqnum = function
         | Some x ->
             Pdfops.Op_BDC (n, Pdf.Name x)
         | None ->
-            Printf.eprintf "Warning: Missing Op_BDC /Properties entry\n%!";
+            Pdfe.log "Warning: Missing Op_BDC /Properties entry\n";
             Pdfops.Op_BDC (n, Pdf.Name p)
       end
   | Pdfops.InlineImage (dict, bytes) ->
@@ -448,8 +449,9 @@ let mkpage getobjnum parent page =
         ([("/Type", Pdf.Name "/Page");
           ("/Parent", Pdf.Indirect parent);
           ("/Resources", page.resources);
-          ("/MediaBox", page.mediabox);
-          ("/Rotate", Pdf.Integer (int_of_rotation page.rotate))]
+          ("/MediaBox", page.mediabox)]
+         @
+          (let i = int_of_rotation page.rotate in if i = 0 then [] else [("/Rotate", Pdf.Integer i)])
       @
         (match page.rest with
          | Pdf.Dictionary d -> d
@@ -461,7 +463,6 @@ let mkpage getobjnum parent page =
 
 (* Build a list of objnum, pdfobject pairs from the ptree. The pages in the
 ptree are just missing their parent entries, so we add those. *)
-(* FIXME: Now we rewrote this for pdf_of_pages with nicer properties, can we get rid of this one? But consider linearization (see below) *)
 let rec objects_of_ptree getobjnum extras = function
   | Lf (pages, parent, this) ->
       let page_objects =
@@ -541,35 +542,17 @@ let rec objects_of_ptree getobjnum extras = function
            in
              this_objects @ objs_left @ objs_right
      
-let flat_pagetrees = ref false
-
-(* Flat version *)
-let add_pagetree_flat pages pdf =
-  let extras = ref [] in
-    let getobjnum = source pdf.Pdf.objects.Pdf.maxobjnum in
-      let ptree = pagetree_flat getobjnum pages 0 in
-        let objects = objects_of_ptree getobjnum extras ptree in
-          let topnode = match hd objects with (n, _) -> n in
-            iter (fun x -> ignore (Pdf.addobj_given_num pdf x)) (objects @ !extras);
-            pdf, topnode
-
 (* Take a list of pages and a PDF. Build a page tree in the PDF, returning
 the new pdf and the object number assigned to the top page node. All references
 to objects not forming part of the tree nodes themselves are left unchanged. *)
 let add_pagetree pages pdf =
-  if !flat_pagetrees then add_pagetree_flat pages pdf else
-    let extras = ref [] in
-      let getobjnum = source pdf.Pdf.objects.Pdf.maxobjnum in
-        let ptree = pagetree getobjnum pages 0 in
-          let objects = objects_of_ptree getobjnum extras ptree in
-            let topnode = match hd objects with (n, _) -> n in
-              (*Printf.printf "There were %i objects_of_ptree\n" (List.length objects);
-              List.iter (fun (i, x) -> Printf.printf "%i: %s\n" i (Pdfwrite.string_of_pdf x)) objects;
-              Printf.printf "There were %i extras\n" (List.length !extras);
-              List.iter (fun (i, x) -> Printf.printf "%i: %s\n" i
-              (Pdfwrite.string_of_pdf x)) !extras;*)
-              iter (fun x -> ignore (Pdf.addobj_given_num pdf x)) (objects @ !extras);
-              pdf, topnode
+  let extras = ref [] in
+    let getobjnum = source pdf.Pdf.objects.Pdf.maxobjnum in
+      let ptree = pagetree getobjnum pages 0 in
+        let objects = objects_of_ptree getobjnum extras ptree in
+          let topnode = match hd objects with (n, _) -> n in
+            iter (fun x -> ignore (Pdf.addobj_given_num pdf x)) (objects @ !extras);
+            pdf, topnode
 
 (* Add a root entry, replacing the Type and Pages entry, and any entries in
 extras. Preserves any entries in any existing root (e.g Metadata pointer). *)
@@ -643,9 +626,9 @@ let change_pages_find_matrix dest mattable refnumstable =
               Hashtbl.find mattable pagenumber
           with
             _ ->
-              Printf.eprintf
-                "page not found for bookmark or annotation dest:%s\n%!"
-                (Pdfwrite.string_of_pdf (Pdfdest.pdfobject_of_destination dest));
+              Pdfe.log (Printf.sprintf
+                "page not found for bookmark or annotation dest:%s\n"
+                (Pdfwrite.string_of_pdf (Pdfdest.pdfobject_of_destination dest)));
               Pdftransform.i_matrix
           end
       | _ -> Pdftransform.i_matrix
@@ -694,7 +677,7 @@ let rewrite_action pdf mattable refnumstable action =
 let change_pages_process_annotations mattable refnumstable pdf =
    iter2
      (fun page pnum ->
-        (*Printf.eprintf "Page %i...\n" pnum;*)
+        (*Pdfe.log (Printf.sprintf "Page %i...\n" pnum);*)
         match Pdf.lookup_direct pdf "/Annots" page.rest with
         | Some (Pdf.Array annots) ->
             iter
@@ -726,11 +709,11 @@ let change_pages_process_annotations mattable refnumstable pdf =
                       end
                   | _ -> ()
                   end
-                | _ -> Printf.eprintf "change_pages_process_annotations: annotation direct\n%!")
+                | _ -> Pdfe.log "change_pages_process_annotations: annotation direct\n")
                 annots
         | None -> ()
         | _ ->
-            Printf.eprintf "change_pages_process_annotations: /Annots not an array\n%!")
+            Pdfe.log "change_pages_process_annotations: /Annots not an array\n")
      (pages_of_pagetree pdf)
      (indx (pages_of_pagetree pdf))
 
@@ -786,7 +769,7 @@ let change_pages ?matrices ?changes change_references basepdf pages' =
                          combine old_page_numbers new_page_numbers
                        else
                          begin
-                           Printf.eprintf "change_pages: No change supplied, and lengths differ\n%!";
+                           Pdfe.log "change_pages: No change supplied, and lengths differ\n";
                            []
                          end
                    | Some cs ->
@@ -813,15 +796,15 @@ let change_pages ?matrices ?changes change_references basepdf pages' =
                           change_pages_process_bookmarks mattable refnumstable pdf
                         else
                           begin
-                            Printf.eprintf "Pdfpage.change_pages: non-null matrices when lengths differ\n%!";
+                            Pdfe.log "Pdfpage.change_pages: non-null matrices when lengths differ\n";
                             pdf
                           end
                       in
                         begin try change_pages_process_annotations mattable refnumstable pdf with
-                          e -> Printf.eprintf "failure in change_pages_process_annotations: %s\n%!" (Printexc.to_string e)
+                          e -> Pdfe.log (Printf.sprintf "failure in change_pages_process_annotations: %s\n" (Printexc.to_string e))
                         end;
                         begin try change_pages_process_openaction mattable refnumstable pdf with
-                          e -> Printf.eprintf "failure in change_pages_process_openaction: %s\n%!" (Printexc.to_string e)
+                          e -> Pdfe.log (Printf.sprintf "failure in change_pages_process_openaction: %s\n" (Printexc.to_string e))
                         end;
                         pdf
 
@@ -829,9 +812,7 @@ let change_pages ?matrices ?changes change_references basepdf pages' =
 same page object numbers, so bookmarks etc still work. Also sorts out bookmarks
 so only those in the range are kept. *)
 
-
 (* Find a page indirect from the page tree of a document, given a page number. *)
-(* FIXME speed up by caching *)
 let page_object_number pdf destpage =
   try
     Some (select destpage (Pdf.page_reference_numbers pdf))
@@ -1096,13 +1077,12 @@ let prepend_operators pdf ops ?(fast=false) page =
         [Pdfops.stream_of_ops (ops @ old_ops)]}
 
 (* Add stack operators to a content stream to ensure it is composeable. *)
-let protect pdf resources content =
-  let ops = Pdfops.parse_operators pdf resources content in
-    let qs = length (keep (eq Pdfops.Op_q) ops)
-    and bigqs = length (keep (eq Pdfops.Op_Q) ops) in
-    let deficit = if qs > bigqs then qs - bigqs else 0 in
-      if deficit <> 0 then Printf.eprintf "Q Deficit was nonzero. Fixing. %i\n%!" deficit;
-      many Pdfops.Op_Q deficit
+let protect ops =
+  let qs = length (keep (eq Pdfops.Op_q) ops)
+  and bigqs = length (keep (eq Pdfops.Op_Q) ops) in
+  let deficit = if qs > bigqs then qs - bigqs else 0 in
+    if deficit <> 0 then Pdfe.log (Printf.sprintf "Q Deficit was nonzero. Fixing. %i\n" deficit);
+    many Pdfops.Op_Q deficit
 
 (* We check for q/Q mismatches in existing section. *)
 let postpend_operators pdf ops ?(fast=false) page =
@@ -1110,13 +1090,15 @@ let postpend_operators pdf ops ?(fast=false) page =
     {page with content =
        page.content @ [Pdfops.stream_of_ops ([Pdfops.Op_q] @ ops @ [Pdfops.Op_Q])]}
   else
-    let beforeops =
-      [Pdfops.Op_q]
-    and afterops =
-      protect pdf page.resources page.content @ [Pdfops.Op_Q] @ ops
-    in
-      {page with content =
-         [Pdfops.stream_of_ops (beforeops @ Pdfops.parse_operators pdf page.resources page.content @ afterops)]}
+    let existing_ops = Pdfops.parse_operators pdf page.resources page.content in
+      if existing_ops = [] then
+        {page with content = [Pdfops.stream_of_ops ops]} (* avoid protecting the empty contents *)
+      else
+        let protected = protect existing_ops in
+        let beforeops = [Pdfops.Op_q] in
+        let afterops = protected @ [Pdfops.Op_Q] @ ops in
+          {page with content =
+             [Pdfops.stream_of_ops (beforeops @ existing_ops @ afterops)]}
 
 (* Source of possible prefix strings. String is always copied. *)
 let next_string s =
@@ -1237,22 +1219,13 @@ let change_resources pdf prefix resources =
 (* For each object in the PDF with /Type /Page or /Type /Pages:
   a) Add the prefix to any name in /Resources
   b) Add the prefix to any name used in any content streams, keeping track of
-  the streams we have processed to preserve sharing
-  
-FIXME: If a non-ISO PDF with content streams which don't end on lexical boundaries
-is provided, the parse will fail, and this function will raise an exception.
-
-The long-term solution to this is to explode the PDF with an
--unshare-content-streams option, and then fix squeeze to re-share subcontent
-streams *)
+  the streams we have processed to preserve sharing *)
 let add_prefix pdf prefix =
   let fixed_streams = Hashtbl.create 100 in
   let fix_stream resources i =
     match i with Pdf.Indirect i ->
-      (*Printf.eprintf "fixing stream %i\n%!" i;*)
       if not (Hashtbl.mem fixed_streams i) then
         let operators = Pdfops.parse_operators pdf resources [Pdf.Indirect i] in
-          (*Printf.eprintf "calling prefix_operator on stream %i\n%!" i;*)
           let operators' = map (prefix_operator pdf prefix) operators in
             begin match Pdf.lookup_obj pdf i with
               Pdf.Stream ({contents = (dict, stream)} as s) ->
@@ -1283,10 +1256,27 @@ let add_prefix pdf prefix =
                        (if resources = None then Pdf.Dictionary [] else unopt resources)
                        (Pdf.Indirect i)
                  | Some (Pdf.Array a) ->
-                     iter
-                       (fix_stream
-                         (if resources = None then Pdf.Dictionary [] else unopt resources))
-                       a
+                     (* May be Non-ISO, and not parse properly individually! If so, detect, and Exit *)
+                     begin match
+                       if length a > 1 then (* Can't be broken when just a single entry *)
+                         iter
+                           (function c ->
+                             let c = Pdf.direct pdf c in
+                               Pdfcodec.decode_pdfstream pdf c;
+                               let s = Pdf.bigarray_of_stream c in
+                               let resources = match resources with Some r -> r | _ -> Pdf.Dictionary [] in
+                                 ignore (Pdfops.parse_single_stream pdf resources s))
+                           a
+                     with
+                     | () ->
+                         iter
+                           (fix_stream
+                             (if resources = None then Pdf.Dictionary [] else unopt resources))
+                           a
+                     | exception _ ->
+                         Pdfe.log "add_prefix: non-ISO PDF detected streams detected. Fixing...\n";
+                         raise Exit
+                     end;
                  | _ -> ()
                  end;
                  begin match resources' with
@@ -1296,5 +1286,121 @@ let add_prefix pdf prefix =
            | _ -> obj
            end
        | _ -> obj)
-    pdf(*;
-    Printf.eprintf "***add_prefix has concluded\n%!";*)
+    pdf
+
+let merge_content_streams pdf = 
+  Pdf.objselfmap
+    (fun obj ->
+       match obj with
+         Pdf.Dictionary dict as d ->
+           begin match Pdf.lookup_direct pdf "/Type" d with
+             Some (Pdf.Name ("/Page" | "/Pages")) ->
+               begin match lookup "/Contents" dict with
+               | Some (Pdf.Indirect _ | Pdf.Array [_]) -> obj
+               | Some (Pdf.Array a) ->
+                   let a = map (Pdf.direct pdf) a in
+                     iter (Pdfcodec.decode_pdfstream pdf) a;
+                     let bigarrays = map Pdf.bigarray_of_stream a in
+                     let merged = Pdfops.concat_bytess bigarrays in
+                     let stream =
+                       Pdf.Stream
+                         {contents =
+                            (Pdf.Dictionary ["/Length", Pdf.Integer (Pdfio.bytes_size merged)],
+                             Pdf.Got merged)}
+                     in
+                     let i = Pdf.addobj pdf stream in
+                       Pdf.add_dict_entry d "/Contents" (Pdf.Indirect i)
+               | _ -> obj
+               end
+           | _ -> obj
+           end
+       | _ -> obj)
+   pdf
+
+(* If a non-ISO PDF with content streams which don't end on lexical boundaries
+is provided, we must merge them. But we don't want to unless we have to,
+because it destroys sharing. *)
+let add_prefix pdf prefix =
+  try add_prefix pdf prefix with Exit ->
+    merge_content_streams pdf;
+    add_prefix pdf prefix
+
+(* For uses of process_pages which don't need to deal with matrices, this
+function transforms into one which returns the identity matrix *)
+let ppstub f n p = (f n p, n, Pdftransform.i_matrix)
+
+let process_xobject f pdf resources i =
+  let xobj = Pdf.lookup_obj pdf i in
+    match Pdf.lookup_direct pdf "/Subtype" xobj with
+    | None -> raise (Pdf.PDFError "No /Subtype in Xobject") 
+    | Some (Pdf.Name "/Form") ->
+        Pdf.getstream xobj;
+        begin match xobj with
+        | Pdf.Stream ({contents = Pdf.Dictionary dict, Pdf.Got bytes} as rf) ->
+            begin match f pdf resources [Pdf.Stream rf] with
+            | [Pdf.Stream {contents = (Pdf.Dictionary dict', data)}] ->
+                let dict' =
+                  Pdf.remove_dict_entry
+                    (Pdf.Dictionary (mergedict dict dict'))
+                    "/Filter"
+                in
+                  rf := (dict', data)
+            | _ -> assert false
+            end
+        | _ -> assert false (* getstream would have complained already *)
+        end
+    | Some _ -> ()
+
+let process_xobjects pdf page f =
+  match Pdf.lookup_direct pdf "/XObject" page.resources with
+  | Some (Pdf.Dictionary elts) ->
+      iter
+        (fun (k, v) ->
+          match v with
+          | Pdf.Indirect i -> process_xobject f pdf page.resources i
+          | _ -> raise (Pdf.PDFError "process_xobject"))
+        elts
+  | _ -> ()
+
+(* Union two resource dictionaries from the same PDF. *)
+let combine_pdf_resources pdf a b =
+  let a_entries =
+    match a with
+    | Pdf.Dictionary entries -> entries
+    | _ -> []
+  in let b_entries =
+    match b with
+    | Pdf.Dictionary entries -> entries
+    | _ -> []
+  in
+    let resource_keys =
+      ["/Font"; "/ExtGState"; "/ColorSpace"; "/Pattern";
+       "/Shading"; "/XObject"; "/Properties"]
+    in
+      let combine_entries key =
+        let a_entries =
+          match Pdf.lookup_direct pdf key a with
+          | Some (Pdf.Dictionary d) -> d
+          | _ -> []
+        in let b_entries =
+          match Pdf.lookup_direct pdf key b with
+          | Some (Pdf.Dictionary d) -> d
+          | _ -> []
+        in
+          if a_entries = [] && b_entries = [] then
+            None
+          else
+            Some (key, Pdf.Dictionary (a_entries @ b_entries))
+      in
+        let unknown_keys_a = lose (fun (k, _) -> mem k resource_keys) a_entries in
+        let unknown_keys_b = lose (fun (k, _) -> mem k resource_keys) b_entries in
+        let combined_known_entries = option_map combine_entries resource_keys in
+          fold_left
+            (fun dict (k, v) -> Pdf.add_dict_entry dict k v)
+            (Pdf.Dictionary [])
+            (unknown_keys_a @ unknown_keys_b @ combined_known_entries)
+
+let minimum_valid_pdf () =
+  let pdf_pages = [blankpage Pdfpaper.a4] in
+  let pdf, pageroot = add_pagetree pdf_pages (Pdf.empty ()) in
+    add_root pageroot [] pdf
