@@ -716,7 +716,7 @@ let rec parse_tounicode pdf tounicode =
                    (lose Pdf.is_whitespace (charlist_of_bytes data))))
           with
             e ->
-              Printf.eprintf "/ToUnicode Parse Error : %s\n%!" (Printexc.to_string e);
+              Pdfe.log (Printf.sprintf "/ToUnicode Parse Error : %s\n" (Printexc.to_string e));
               raise e
           end
       | _ -> assert false
@@ -737,7 +737,7 @@ let add_tounicode pdf font fontdict =
         begin try
           Some (hashtable_of_dictionary <| parse_tounicode pdf tounicode)
         with
-          e -> Printf.eprintf "bad tounicode (%s)\n%!" (Printexc.to_string e); None
+          e -> Pdfe.log (Printf.sprintf "bad tounicode (%s)\n" (Printexc.to_string e)); None
         end
     | None -> None
   in
@@ -843,7 +843,7 @@ let tounicode_map s us =
 
 (* Just for the kind produced by the font subsetter for now. *)
 let write_tounicode pdf u =
-  let bytes = tounicode_map 0 (map (fun (_, s) -> s) (sort compare ((list_of_hashtbl u)))) in
+  let bytes = tounicode_map 33 (map (fun (_, s) -> s) (sort compare ((list_of_hashtbl u)))) in
     Pdf.addobj pdf (Pdf.Stream {contents = (Pdf.Dictionary [("/Length", Pdf.Integer (bytes_size bytes))], Pdf.Got bytes)})
 
 let write_font ?objnum pdf = function
@@ -1109,7 +1109,7 @@ let charcode_extractor_of_font_real ?(debug=false) font =
             in
             hashtable_of_dictionary reversed, true
         with
-          e -> Printf.eprintf "bad tounicode (%s)\n%!" (Printexc.to_string e); (null_hash (), false)
+          e -> Pdfe.log (Printf.sprintf "bad tounicode (%s)\n" (Printexc.to_string e)); (null_hash (), false)
         end
     | None -> null_hash (), false
   in
@@ -1126,7 +1126,7 @@ let charcode_extractor_of_font_real ?(debug=false) font =
             Some (let r = Hashtbl.find table glyphname in if debug then Printf.printf "Found charcode %i\n\n" r; r)
       with
         Not_found ->
-          if debug then Printf.eprintf "Found no charcode for unicode codepoint %X.\n" codepoint;
+          if debug then Pdfe.log (Printf.sprintf "Found no charcode for unicode codepoint %X.\n" codepoint);
           None
 
 let charcode_extractor_of_font ?(debug=false) pdf fontdict =
@@ -1145,10 +1145,10 @@ let codepoint_of_pdfdocencoding_character i =
         Hashtbl.find (Pdfglyphlist.glyph_hashes ())
           (Hashtbl.find Pdfglyphlist.reverse_name_to_pdf_hashes i)
       with
-      | [codepoint] -> Some codepoint
-      | _ -> Printf.eprintf "codepoint_of_pdfdocencoding: bad text string (char %i)\n%!" i; None
+      | [codepoint] -> codepoint
+      | _ -> raise (Pdf.PDFError (Printf.sprintf "codepoint_of_pdfdocencoding: bad text string (char %i)\n" i))
     with
-      _ -> Printf.eprintf "codepoint_of_pdfdocencoding: bad text string (char %i)\n%!" i; None
+      _ -> raise (Pdf.PDFError (Printf.sprintf "codepoint_of_pdfdocencoding: bad text string 2 (char %i)\n" i))
 
 (* Build a UTF-8 string from a list of unicode codepoints. *)
 let get_utf8_chars c =
@@ -1174,11 +1174,11 @@ let codepoints_of_pdfdocstring s =
   if is_unicode s then
     codepoints_of_utf16be (String.sub s 2 (String.length s - 2))
   else
-    option_map codepoint_of_pdfdocencoding_character (map int_of_char (explode s))
+    map codepoint_of_pdfdocencoding_character (map int_of_char (explode s))
 
 let utf8_of_pdfdocstring s =
-  try utf8_of_codepoints (codepoints_of_pdfdocstring s) with
-    e -> Printf.eprintf "utf8_of_pdfdocstring : %s\n%!" (Printexc.to_string e); ""
+  (*Printf.printf "utf8_of_pdfdocstring: Pdf string is %S\n" s;*)
+  utf8_of_codepoints (codepoints_of_pdfdocstring s)
 
 (* Build a PDFDocEncoding or UTF16BE string from a UTF8 encoded string *)
 let rec codepoints_of_utf8 = function
@@ -1202,7 +1202,7 @@ let rec codepoints_of_utf8 = function
         lor ((c3 land 0b00_11_11_11) lsl 6)
         lor (c4 land 0b00_11_11_11)::codepoints_of_utf8 cs
   | _ ->
-      Printf.eprintf "Bad UTF8 in codepoints_of_utf8\n%!"; []
+      raise (Pdf.PDFError "Bad UTF8 in codepoints_of_utf8\n")
 
 let codepoints_of_utf8 s = codepoints_of_utf8 (map int_of_char (explode s))
 
@@ -1225,6 +1225,7 @@ let pdfdocstring_of_codepoints codepoints =
     Exit -> utf16be_of_codepoints codepoints
 
 let pdfdocstring_of_utf8 s =
+  (*Printf.printf "utf8_of_pdfdocstring: JSON string is %S\n" s;*)
   pdfdocstring_of_codepoints (codepoints_of_utf8 s)
 
 (* PDF strings (except /ID in the trailer dictionary and inside page content
