@@ -199,6 +199,23 @@ let rev_map5 f a b c d e =
 let map5 f a b c d e =
   rev (rev_map5 f a b c d e)
 
+let rec rev_map6_inner f a b c d e g outputs =
+  match a, b, c, d, e, g with
+  | [], [], [], [], [], [] -> outputs
+  | ha::ta, hb::tb, hc::tc, hd::td, he::te, hg::tg ->
+      rev_map6_inner f ta tb tc td te tg (f ha hb hc hd he hg::outputs)
+  | _ -> raise (Invalid_argument "Pdfutil.map6")
+
+let rev_map6 f a b c d e g =
+  rev_map6_inner f a b c d e g []
+
+let map6 f a b c d e g =
+  rev (rev_map6 f a b c d e g)
+
+let sum = fold_left ( + ) 0
+
+let fsum = fold_left ( +. ) 0.
+
 (* Calculate the cumulative sum of a list given a base e.g [cumulative_sum 5
 [1;2;3] = [6; 8; 11]] *)
 let cumulative_sum b l = 
@@ -284,6 +301,8 @@ let i32mul = Int32.mul
 
 let i32div = Int32.div
 
+let sr32 = Int32.shift_right
+
 let lsr32 = Int32.shift_right_logical
 
 let lsl32 = Int32.shift_left
@@ -318,6 +337,8 @@ let i64mul = Int64.mul
 
 let i64div = Int64.div
 
+let sr64 = Int64.shift_right
+
 let lsr64 = Int64.shift_right_logical
 
 let lsl64 = Int64.shift_left
@@ -1276,5 +1297,14 @@ let time = ref 0. (*ref (clock ())*)
 let tt' () = ()
   (*(*Gc.major ();*)
   let t = clock () in
-    Printf.eprintf "Elapsed: %.2f\n%!" (t -. !time);
+    Pdfe.log (Printf.sprintf "Elapsed: %.2f\n" (t -. !time));
     time := t*)
+
+let starts_with prefix s =
+  let len_s = String.length s
+  and len_pre = String.length prefix in
+  let rec aux i =
+    if i = len_pre then true
+    else if String.get s i <> String.get prefix i then false
+    else aux (i + 1)
+  in len_s >= len_pre && aux 0
