@@ -1,5 +1,6 @@
-(** General Functions. Typically one will [open Pdfutil]. *)
+(** General functions
 
+    Typically one will [open Pdfutil]. *)
 
 (** {2 Debug Printing} *)
 
@@ -45,6 +46,9 @@ val string_replace_all : string -> string -> string -> string
 (** The same, but provide a function for the replacement string *)
 val string_replace_all_lazy : string -> (unit -> string) -> string -> string
 
+(** The standard OCaml [starts_with] function, for versions of OCaml too old to have it. *)
+val starts_with : string -> string -> bool
+
 (** {2 Lists} *)
 
 val hd : ('a list -> 'a)
@@ -75,6 +79,12 @@ val sort : (('a -> 'a -> int) -> 'a list -> 'a list)
 (** Tail-recursive versions of list functions (and some simple variations). See
 [Stdlib] for documentation. *)
 
+(** Sum of a list of integers. *)
+val sum : int list -> int
+
+(** Sum of a list of floats. *)
+val fsum : float list -> float
+
 (** Cumulative sum of a list given an initial value. For example, [cumulative_sum 1 [2; 3; 4]] is [[3; 6; 10]] *)
 val cumulative_sum : int -> int list -> int list
 
@@ -126,6 +136,16 @@ val map5 :
   ('a -> 'b -> 'c -> 'd -> 'e -> 'f) ->
   'a list -> 'b list -> 'c list -> 'd list -> 'e list -> 'f list
 
+(** Similar to [rev_map], but 6 arguments. *)
+val rev_map6 :
+  ('a -> 'b -> 'c -> 'd -> 'e -> 'f -> 'g) ->
+  'a list -> 'b list -> 'c list -> 'd list -> 'e list -> 'f list -> 'g list
+
+(** Similar to [map2], but 6 arguments. *)
+val map6 :
+  ('a -> 'b -> 'c -> 'd -> 'e -> 'f -> 'g) ->
+  'a list -> 'b list -> 'c list -> 'd list -> 'e list -> 'f list -> 'g list
+
 (** [take l n] takes [n] elements from the list raising [Invalid_argument] if
 there are not enough elements to take or if n < 0. *)
 val take : 'a list -> int -> 'a list
@@ -344,6 +364,7 @@ val i32add : (int32 -> int32 -> int32)
 val i32sub : (int32 -> int32 -> int32)
 val i32div : (int32 -> int32 -> int32)
 val i32mul : (int32 -> int32 -> int32)
+val sr32 : (int32 -> int -> int32)
 val lsr32 : (int32 -> int -> int32)
 val lsl32 : (int32 -> int -> int32)
 val lor32 : (int32 -> int32 -> int32)
@@ -362,6 +383,7 @@ val i64add : (int64 -> int64 -> int64)
 val i64sub : (int64 -> int64 -> int64)
 val i64div : (int64 -> int64 -> int64)
 val i64mul : (int64 -> int64 -> int64)
+val sr64 : (int64 -> int -> int64)
 val lsr64 : (int64 -> int -> int64)
 val lsl64 : (int64 -> int -> int64)
 val lor64 : (int64 -> int64 -> int64)
